@@ -16,6 +16,7 @@
 
 import re
 import sys
+import os
 from os import path
 import time
 import datetime
@@ -26,27 +27,31 @@ from calculate.lib.cl_xml import ET
 import pexpect
 from calculate.lib.configparser import ConfigParser
 
-from calculate.lib.utils.colortext.palette import TextState
-from calculate.lib.utils.common import cmpVersion
-from calculate.lib.utils.files import (getProgPath, find, process,
-                                       readLinesFile, pathJoin)
-from calculate.lib.utils.tools import SavableIterator, ignore
+from .colortext.palette import TextState
+from .common import cmpVersion
+from .files import (getProgPath, find, process, listDirectory, readFile,
+                    readLinesFile, pathJoin, makeDirectory,
+                    FilesError, rsync_files, RsyncOptions,
+                    removeDir, removeFileWithEmptyDirectory)
+from .tools import SavableIterator, ignore
 from collections import Mapping, defaultdict
-from common import getTupleVersion
+from .common import getTupleVersion
 from contextlib import closing
-from files import listDirectory, readFile
 from functools import total_ordering
 from itertools import ifilter, imap, chain
 
 Colors = TextState.Colors
 import glob
 
+_ = lambda x: x
 from calculate.lib.cl_lang import setLocalTranslate
 
 setLocalTranslate('cl_lib3', sys.modules[__name__])
 
-reVerSplit = re.compile(r"^(?:.*/var/db/pkg/)?(?:(\w+-\w+)/)?(.*?)-(([^-]+?)"
-                        "(?:-(r\d+))?)(?:.(tbz2))?$", re.S)
+VDB_PATH = 'var/db/pkg'
+
+reVerSplit = re.compile(r"^(?:.*/%s/)?(?:(\w+-\w+)/)?(.*?)-(([^-]+?)"
+                        "(?:-(r\d+))?)(?:.(tbz2))?$" % VDB_PATH, re.S)
 
 class RepositoryPath(object):
     Layman = '/var/lib/layman'
@@ -80,7 +85,7 @@ def reVerSplitToPV(x):
 def getPkgUses(fullpkg, version=None, prefix="/"):
     """Get USE and IUSE from package"""
     category, slash, pkg = fullpkg.partition('/')
-    _pkgCategory = 'var/db/pkg/{0}'.format(category)
+    _pkgCategory = '{0}/{1}'.format(VDB_PATH, category)
     pkgCategory = path.join(prefix, _pkgCategory)
     if version is None:
         packages = filter(lambda x: x['PN'] == pkg,
@@ -107,7 +112,7 @@ def getPkgUses(fullpkg, version=None, prefix="/"):
 
 def isPkgInstalled(pkg, prefix='/', sortByVersion=False):
     """Check is package installed"""
-    pkgDir = path.join(prefix, 'var/db/pkg')
+    pkgDir = path.join(prefix, VDB_PATH)
     if "/" in pkg:
         category, op, pkgname = pkg.partition('/')
 
@@ -133,7 +138,7 @@ def isPkgInstalled(pkg, prefix='/', sortByVersion=False):
 def getPkgSlot(pkg, prefix='/'):
     """Get package slot"""
     pkgs = isPkgInstalled(pkg, prefix)
-    pkgDir = path.join(prefix, 'var/db/pkg')
+    pkgDir = path.join(prefix, VDB_PATH)
     return [readFile(path.join(pkgDir, x['CATEGORY'],
                                x['PF'], "SLOT")).strip()
             for x in pkgs]
@@ -147,23 +152,47 @@ def getPkgActiveUses(fullpkg):
     return list(set(res[0]) & set(res[1]))
 
 
-def getInstalledAtom(atom, prefix='/'):
-    atom = EmergePackage(atom)
-    db_path = path.join(prefix, 'var/db/pkg')
-    glob_fn = path.join(db_path, "%s-*" % atom["CATEGORY/PN"])
-    for pkg_dn in glob.glob(glob_fn):
+def getInstalledAtom(str_atom, prefix='/'):
+    """
+    Получить установленные пакеты по указанному ATOM
+    :param atom:
+    :param prefix:
+    :return:
+    """
+    useglob = "*" in str_atom
+    atom = EmergePackage(str_atom)
+    db_path = path.join(prefix, VDB_PATH)
+
+    def get_package_dn():
+        if useglob:
+            if str_atom == "*":
+                mask = "*/*"
+            else:
+                mask= str_atom
+            glob_fn = path.join(db_path, "%s/*.ebuild" % mask)
+            for pkg_dn in (path.dirname(x) for x in glob.glob(glob_fn)):
+                yield pkg_dn
+        else:
+            glob_fn = path.join(db_path, "%s-*" % atom["CATEGORY/PN"])
+            for pkg_dn in glob.glob(glob_fn):
+                yield pkg_dn
+
+    for pkg_dn in get_package_dn():
         slot = readFile(path.join(pkg_dn, "SLOT"))
         # удаляем subslot
         slot = slot.partition("/")[0]
         if not slot.strip():
             continue
         find_atom = EmergePackage("%s:%s" % (pkg_dn[len(db_path) + 1:], slot))
-        if find_atom["CATEGORY/PN"] == atom["CATEGORY/PN"]:
-            if atom["SLOT!"] and atom["SLOT"] != find_atom["SLOTONLY"]:
-                continue
-            if atom["PN"] != atom["PF"] and find_atom["PF"] != atom["PF"]:
-                continue
+        if useglob:
             yield find_atom
+        else:
+            if find_atom["CATEGORY/PN"] == atom["CATEGORY/PN"]:
+                if atom["SLOT!"] and atom["SLOT"] != find_atom["SLOTONLY"]:
+                    continue
+                if atom["PN"] != atom["PF"] and find_atom["PF"] != atom["PF"]:
+                    continue
+                yield find_atom
 
 
 def getSquashList():
@@ -447,13 +476,13 @@ class EmergePackage(Mapping):
     по версии
     """
     default_repo = 'gentoo'
-    prefix = r"(?:.*/var/db/pkg/|=)?"
+    prefix = r"(?:.*/%s/|=)?" % VDB_PATH
     category = r"(?:(\w+(?:-\w+)?)/)?"
-    pn = "([^/]*?)"
-    pv = r"(?:-(\d[^-]*?))?"
+    pn = "([^/:]*?)"
+    pv = r"(?:-(\d[^:-]*?))?"
     pr = r"(?:-(r\d+))?"
     tbz = r"(?:.(tbz2))?"
-    slot = r'(?::(\w+(?:\.\w+)*(?:/\w+(?:\.\w+)*(?:-\w+(?:\.\w+)*)?)?))?'
+    slot = r'(?::([^:/]+(?:/[^:/]+)?))?'
     repo = r'(?:::(\w+))?'
 
     reParse = re.compile(
@@ -462,7 +491,7 @@ class EmergePackage(Mapping):
             slot=slot, repo=repo))
 
     attrs = ('CATEGORY', 'PN', 'PF', 'SLOT', 'REPO', 'P', 'PV', 'PR', 'PVR',
-             'CATEGORY/PN', 'SLOT!', 'SLOTONLY')
+             'CATEGORY/PN', 'SLOT!', 'SLOTONLY', 'CATEGORY/PF')
 
     def _parsePackageString(self, s):
         """
@@ -489,6 +518,7 @@ class EmergePackage(Mapping):
                 d['PVR'] = d['PV']
             if d['PF'].endswith('-r0'):
                 d['PF'] = d['PF'][:-3]
+            d['CATEGORY/PF'] = "%s/%s" % (x[CATEGORY], d['PF'])
             return d.copy()
         else:
             return {k: '' for k in self.attrs}
@@ -540,7 +570,10 @@ class EmergePackage(Mapping):
         return "EmergePackage(%s/%s)" % (self['CATEGORY'], self['PF'])
 
     def __str__(self):
-        return "%s/%s" % (self['CATEGORY'], self['PF'])
+        return self['CATEGORY/PF']
+
+    def __hash__(self):
+        return hash(str(self))
 
 
 class PackageInformation:
@@ -1182,10 +1215,10 @@ class InstalledPackageInfo(object):
         return self._info[item]
 
     @classmethod
-    def get_install_packages(cls, pkg_dir='/var/db/pkg'):
+    def get_install_packages(cls, pkg_dir='/%s' % VDB_PATH):
         for category in listDirectory(pkg_dir):
             for pkg in listDirectory(path.join(pkg_dir, category)):
-                if "MERGING" not in pkg:
+                if "MERGING" not in pkg and "lockfile" not in pkg:
                     yield InstalledPackageInfo("%s/%s" % (category, pkg),
                                                pkg_dir=pkg_dir)
 
@@ -1206,3 +1239,197 @@ def makeCfgName(origfilename):
         if not path.exists(fn):
             return fn
     return origfilename
+
+def getlibpaths(prefix="/"):
+    """
+    Возвращает список путей
+    """
+    ld_so_include_re = re.compile(r'^include\s+(\S.*)')
+
+    def read_ld_so_conf(fn):
+        for l in readLinesFile(fn, grab=True):
+            include_match = ld_so_include_re.match(l)
+            if include_match is not None:
+                subpath = path.join(path.dirname(fn),
+                                    include_match.group(1))
+                for p in glob.glob(subpath):
+                    for r in read_ld_so_conf(p):
+                        yield r
+            else:
+                yield l
+    rval = list(read_ld_so_conf(path.join(prefix, "etc", "ld.so.conf")))
+    rval.append("/usr/lib")
+    rval.append("/lib")
+    return sorted(list(set([path.normpath(x) for x in rval if x])))
+
+_re_req_arches = re.compile("(\w+:\s.*?)(?=\s\w+:|$)")
+
+
+class LibraryProviders(Mapping):
+    """
+    Пакеты, которым принадлежат библиотеки
+    TODO: sync to portage.py
+    """
+
+    def __init__(self, vdb_path=VDB_PATH, prefix="/"):
+        self.data = {}
+        vdb = path.join(prefix, vdb_path)
+        lvdb = len(vdb) + 1
+        libpath = getlibpaths(prefix)
+        pattern_libpath = "|".join(re.escape(x) for x in libpath)
+
+        for provides in glob.glob(path.join(vdb, "*/*/PROVIDES"))[1:]:
+            p = path.dirname(provides[lvdb:])
+            slot = readFile(provides.replace(
+                "PROVIDES", "SLOT")).strip().partition('/')[0]
+            package = EmergePackage("%s:%s" % (p, slot))
+            content = readFile(provides.replace(
+                "PROVIDES", "CONTENTS"))
+
+            for arch, libraries in (
+                    x.partition(": ")[::2]
+                    for x in _re_req_arches.findall(readFile(provides))):
+                if arch not in self.data:
+                    self.data[arch] = defaultdict(list)
+                for library in list(set(
+                        re.findall("^(?:sym|obj)\s+(?:%s)/(%s)\s+" % (
+                            pattern_libpath, "|".join(
+                                re.escape(x) for x in libraries.split())),
+                               content, flags=re.M))):
+                    self.data[arch][library].append(package)
+
+    def __getitem__(self, item):
+        return self.data[item]
+
+    def __iter__(self):
+        return iter(self.data)
+
+    def __len__(self):
+        return len(self.data)
+
+
+def getRequires(package, vdb_path=VDB_PATH, prefix='/'):
+    """
+    Получить список библиотек требуемых для работы пакета
+    :param package:
+    :param prefix:
+    :return:
+    """
+    package = EmergePackage(package)
+    vdb = path.join(prefix, vdb_path)
+    requires = "%s/%s/REQUIRES"%(vdb, package["CATEGORY/PF"])
+    for arch, libraries in (x.partition(": ")[::2]
+                            for x in _re_req_arches.findall(readFile(requires))):
+        for require in libraries.split():
+            yield arch, require
+
+class PackageError(Exception):
+    pass
+
+def hide_packages(*packages, **kw):
+    """
+    Спрятать пакет (для исключение автозависимостей при сборки некоторых
+    пакетов) prefix (/), force_remove=False
+    :param package:
+    :param prefix:
+    :return:
+    """
+    prefix = kw.get('prefix', '/')
+    force_remove = kw.get('force_remove', False)
+    from .content import PkgContents
+    base_dn = path.join(prefix, 'var/lib/calculate/calculate-builder')
+    hide_dn = path.join(base_dn, "automagic-requires")
+
+    if path.exists(hide_dn):
+        if force_remove:
+            removeDir(hide_dn)
+        else:
+            raise PackageError(
+                _("Failed to hide {packages}: {error}").format(
+                    package=",".join(str(x) for x in packages),
+                    error=_("package is already hidden")))
+
+    makeDirectory(hide_dn)
+
+    def arch_list():
+        for package in packages:
+            content = PkgContents(package, prefix=prefix)
+            for fn, obj_data in content.content.items():
+                if obj_data['type'] in ("obj", "sym"):
+                    yield fn
+
+    arch_files = list(arch_list())
+    try:
+        arch_files = [x for x in arch_files
+                      if path.lexists(pathJoin(prefix, x))]
+        rsync_files(prefix, hide_dn, *arch_files,
+                    opts=RsyncOptions.HidePackageOpts)
+        for fn in (pathJoin(prefix, x) for x in arch_files):
+            os.unlink(fn)
+    except FilesError as e:
+        raise PackageError(
+            _("Failed to hide {package}: {error}".format(
+                package=",".join(str(x) for x in packages),
+                error=str(e))))
+
+def unhide_packages(prefix='/', force=False):
+    """
+    Восстановить спрятанные файлы пакета, prefix (/)
+    :param package:
+    :param prefix:
+    :return:
+    """
+    # /var/lib/calculate/calculate-builder/automagic-requires
+    base_dn = path.join(prefix, 'var/lib/calculate/calculate-builder')
+    hide_dn = path.join(base_dn, "automagic-requires")
+
+    if not path.exists(hide_dn):
+        if not force:
+            raise PackageError(
+                _("Failed to unhide packages: {error}").format(
+                    error=_("directory not found")))
+        return
+
+    try:
+        rsync_files(hide_dn, prefix, opts=RsyncOptions.HidePackageOpts)
+        removeDir(hide_dn)
+        removeFileWithEmptyDirectory(hide_dn, stopDirectory=base_dn)
+    except FilesError as e:
+        raise PackageError(
+            _("Failed to unhide: {error}").format(
+                error=str(e)))
+
+XPAK_TMP = "/dev/shm/xpak"
+
+class BinaryPackage(object):
+    """
+    Объект для модификации var/db файлов бинарного пакета
+    """
+    def __init__(self, xpak_fn, work_dn=XPAK_TMP):
+        from calculate.lib.cl_portage import xpak
+        self.work_dn = work_dn
+        self.tbz2 = xpak.tbz2(xpak_fn)
+        self.tbz2.decompose(self.work_dn)
+
+    def __setitem__(self, item, value):
+        fn = path.join(self.work_dn, item)
+        open(fn, 'w').write("%s\n"%value.rstrip())
+
+    def __getitem__(self, item):
+        fn = path.join(self.work_dn, item)
+        return open(fn, 'r').read().strip()
+
+    def save(self):
+        self.tbz2.compose(self.work_dn)
+
+    def clear(self):
+        removeDir(self.work_dn)
+
+def get_binary_file(pkg, pkgdir):
+    """
+    Получить имя файла бинарного пакета
+    :param pkg:
+    :param pkgdir:
+    :return:
+    """
+    return pathJoin(pkgdir, "%s.tbz2"%str(pkg))
